// Global const
const CALIBRATION_START_PREFIX = "calibration started: ";
const CALIBRATION_COMPLETE_PREFIX = "calibration done: ";
const CALIBRATION_PROGRESS_PREFIX = "progress";
const CALIBRATION_CANCEL_PREFIX = "calibration failed";
const CALIBRATION_FAIL_PREFIX = "calibration cancelled";

let calibrationStatus = document.getElementById("calibration-status");
let orientationCalAreaHelpText = document.getElementById("orientationCalAreaHelpText");
let progressCalPercentText = document.getElementById("progressPercentText");


// Calibration Images
let accelDownImage = document.getElementById("accel_down")
let accelUpImage = document.getElementById("accel_up")
let accelFrontImage = document.getElementById("accel_front")
let accelBackImage = document.getElementById("accel_back")
let accelLeftImage = document.getElementById("accel_left")
let accelRightImage = document.getElementById("accel_right")


let startCalibrationButton = document.getElementById("start_calibration_button")
let cancelCalibrationButton = document.getElementById("cancel_calibration_button")

// accelDownArrow.style.display = 'none';

let calibrationTypeInProgress = null;
let accelCalInProgress = false;
let gyroCalInProgress = false;
let magCalInProgress = false;
let levelCalInProgress = false;

let sensor = document.getElementById("sensor-select");
sensor.addEventListener('input', () => {
    if (calibrationTypeInProgress) {
        console.log('sensor val changed, cancelling calibration');
        cancelCalibration();
    }
});

setCalibrationImageVisibility(sensor.value);


sensor.addEventListener('change', function() {
    setCalibrationImageVisibility(sensor.value);
  }
);

function setCalibrationImageVisibility(sensorVal) {
    if (sensorVal == "Gyro" || sensorVal == "Level") {
        accelUpImage.style.display = 'none';
        accelFrontImage.style.display = 'none';
        accelBackImage.style.display = 'none';
        accelLeftImage.style.display = 'none';
        accelRightImage.style.display = 'none';
    } else {
        accelUpImage.style.display = '';
        accelFrontImage.style.display = '';
        accelBackImage.style.display = '';
        accelLeftImage.style.display = '';
        accelRightImage.style.display = '';
    }

    let accelDownArrow = document.getElementById("accel_down_arrow");
    let accelUpArrow = document.getElementById("accel_up_arrow");
    let accelFrontArrow = document.getElementById("accel_front_arrow");
    let accelBackArrow = document.getElementById("accel_back_arrow");
    let accelLeftArrow = document.getElementById("accel_left_arrow");
    let accelRightArrow = document.getElementById("accel_right_arrow");

    accelDownArrow.style.display = 'none';
    accelUpArrow.style.display = 'none';
    accelFrontArrow.style.display = 'none';
    accelBackArrow.style.display = 'none';
    accelLeftArrow.style.display = 'none';
    accelRightArrow.style.display = 'none';

    accelDownImage.style.borderColor = "red";
    accelUpImage.style.borderColor = "red";
    accelLeftImage.style.borderColor = "red";
    accelRightImage.style.borderColor = "red";
    accelFrontImage.style.borderColor = "red";
    accelBackImage.style.borderColor = "red";
};


function calibrateSensor() {
    let sensor = document.getElementById("sensor-select").value;

    fetch('/_cmd/mag_calibration_start?sensor=' + sensor).then(response => {
        // spinner.style.display = "none";
    });
}

async function disableButtons() {
    alert("Waiting for calibration to cancel...");
    startCalibrationButton.disabled = true;
    cancelCalibrationButton.disabled = true;

    while (calibrationTypeInProgress) {
        await new Promise(r => setTimeout(r, 2000));
    }
    startCalibrationButton.disabled = false;
    cancelCalibrationButton.disabled = false;

}

function cancelCalibration() {

    fetch('/_cmd/mag_calibration_stop').then(response => {
        // spinner.style.display = "none";
    });
    disableButtons();
}

// Helper functions
function onCalibrationMessage(text) {
    if (!text.startsWith(CALIBRATION_START_PREFIX) && (calibrationTypeInProgress == null)) {
        console.log("invalid message: " + text);
        return;
    }

    if (calibrationTypeInProgress == null) {

        calibrationStatus.innerHTML = "Calibration Started" 
        text = text.slice(CALIBRATION_START_PREFIX.length);
        // Parse the calibration type: e.g. accel, gyro, mag from the MAVLINK_MSG_STATUS_TEXT
        parts = text.split(" ");
        text = parts[1];
        console.log("part " + text);

        calibrationTypeInProgress = text;
    
        // Reset logic should be here
        if (text == "accel" || text == "mag" || text == "gyro" || text == "level") {
            console.log("Valid sensor for calibration");
            if(text == "accel") {
                accelCalInProgress = true;
            } else if(text == "mag") {
                magCalInProgress = true;
            } else if(text == "gyro") {
                gyroCalInProgress = true;
            } else if (text == "level") {
                levelCalInProgress = true;
            }

            progressCalPercentText.innerHTML = "Calibration Progress: <0>%"; 
            updateOrientationCalArea("Place your vehicle into one of the incomplete orientations shown below and hold it still");
        }

    } else if (calibrationTypeInProgress && (accelCalInProgress || magCalInProgress || gyroCalInProgress || levelCalInProgress)) {

   
        calibrationStatus.innerHTML = "Calibration in Progress" 
        //TODO replace with switch
        //We are in the correct orientation
        if (text.endsWith("orientation detected")) {
            handleOrientationDetected(text.split(" ")[0], magCalInProgress);
            return;
        }
        // Switch to the next side
        if (text.endsWith("side done, rotate to a different side")) {
            updateOrientationCalArea("Side done: place your vehicle into one of the incomplete orientations shown below and hold it still");
            handleSideDone(text.split(" ")[0], magCalInProgress);
            return;
        }
        // Already completed
        if (text.endsWith("side already completed")) {
            updateOrientationCalArea("Orientation previously completed: place your vehicle into one of the incomplete orientations shown below and hold it still");
            return;
        }

        // Update progress
        if (text.startsWith(CALIBRATION_PROGRESS_PREFIX)) {
            progress_elem = text.split(" ");
            progress = progress_elem[1];

            if (gyroCalInProgress || levelCalInProgress) {
                accelDownImage.style.borderColor = "yellow";
            }


            progressCalPercentText.innerHTML = "Calibration Progress: " + progress + "%"; 
            console.log("Progress: " + progress);
        }
    
        // Calibration complete!
        if (text.startsWith(CALIBRATION_COMPLETE_PREFIX)) {
            if (gyroCalInProgress || levelCalInProgress) {
                accelDownImage.style.borderColor = "green";
            }

            stopCalibration("Calibration success");
            return;
        }

        if (text.startsWith(CALIBRATION_CANCEL_PREFIX)) {
            setCalibrationImageVisibility(sensor.value);
            stopCalibration("Calibration cancelled");
            return;

        }

        if (text.startsWith(CALIBRATION_FAIL_PREFIX)) {
            setCalibrationImageVisibility(sensor.value);
            stopCalibration("Calibration failed");
            return;
        }

    }

    else {

        console.log("Unknown state");
    }
}

function handleOrientationDetected(side, magCalInProgress) {
    if (magCalInProgress) {
        updateOrientationCalArea("Rotate the vehicle continuously until marked complete");

    }
    else {
        updateOrientationCalArea("Hold still in the current orientation");
    }

    switch (side) {
        case "down":
            orientationCalDownSideInProgress = true;
            accelDownImage.style.borderColor = "yellow";
            if (magCalInProgress) {
                let accelDownArrow = document.getElementById("accel_down_arrow");
                accelDownArrow.style.display = 'inline-block';
             }
            break;
        case "up":
            orientationCalUpsideDownSideInProgress = true;
            accelUpImage.style.borderColor = "yellow";
            if (magCalInProgress) {
                let accelUpArrow = document.getElementById("accel_up_arrow");
                accelUpArrow.style.display = 'inline-block';
             }
            break;
        case "left":
            orientationCalLeftSideInProgress = true;
            accelLeftImage.style.borderColor = "yellow";
            if (magCalInProgress) {
                let accelLeftArrow = document.getElementById("accel_left_arrow");
                accelLeftArrow.style.display = 'inline-block';
             }
            break;
        case "right":
            orientationCalRightSideInProgress = true;
            accelRightImage.style.borderColor = "yellow";
            if (magCalInProgress) {
                let accelRightArrow = document.getElementById("accel_right_arrow");
                accelRightArrow.style.display = 'inline-block';
             }
            break;
        case "front":
            orientationCalNoseDownSideInProgress = true;
            accelFrontImage.style.borderColor = "yellow";
            if (magCalInProgress) {
                console.log('setting front arrow');
                let accelFrontArrow = document.getElementById("accel_front_arrow");
                accelFrontArrow.style.display = 'inline-block';
             }
            break;
        case "back":
            orientationCalTailDownSideInProgress = true;
            accelBackImage.style.borderColor = "yellow";
            if (magCalInProgress) {
                let accelBackArrow = document.getElementById("accel_back_arrow");
                accelBackArrow.style.display = 'inline-block';
             }
            break;

    }

}

function resetCalibrationState() {
    orientationCalInProgress = false;
    calibrationTypeInProgress = null;
    // Reset all in progress sensors
    magCalInProgress = false;
    accelCalInProgress = false;
    gyroCalInProgress = false;

    // Reset the orientation cal text
    updateOrientationCalArea("");

    // Reset the progress percent
    progressCalPercentText.innerHTML = ""; 
}


function handleSideDone(side) {
    switch (side) {
        case "down":
            orientationCalDownSideInProgress = false;
            orientationCalDownSideDone = true;
            accelDownImage.style.borderColor = "green";
            if (magCalInProgress) {
                let accelDownArrow = document.getElementById("accel_down_arrow");
                accelDownArrow.style.display = 'none';
             }
            break;
        case "up":
            orientationCalUpsideDownSideInProgress = false;
            orientationCalUpsideDownSideDone = true;
            accelUpImage.style.borderColor = "green";
            if (magCalInProgress) {
                let accelUpArrow = document.getElementById("accel_up_arrow");
                accelUpArrow.style.display = 'none';
             }
            break;
        case "left":
            orientationCalLeftSideInProgress = false;
            orientationCalLeftSideDone = true;
            accelLeftImage.style.borderColor = "green";
            if (magCalInProgress) {
                let accelLeftArrow = document.getElementById("accel_left_arrow");
                accelLeftArrow.style.display = 'none';
             }
            break;
        case "right":
            orientationCalRightSideInProgress = false;
            orientationCalRightSideDone = true;
            accelRightImage.style.borderColor = "green";
            if (magCalInProgress) {
                let accelRightArrow = document.getElementById("accel_right_arrow");
                accelRightArrow.style.display = 'none';
             }
            break;
        case "front":
            orientationCalNoseDownSideInProgress = false;
            orientationCalNoseDownSideDone = true;
            accelFrontImage.style.borderColor = "green";
            if (magCalInProgress) {
                let accelFrontArrow = document.getElementById("accel_front_arrow");
                accelFrontArrow.style.display = 'none';
             }
            break;
        case "back":
            orientationCalTailDownSideInProgress = false;
            orientationCalTailDownSideDone = true;
            accelBackImage.style.borderColor = "green";
            if (magCalInProgress) {
                let accelBackArrow = document.getElementById("accel_back_arrow");
                accelBackArrow.style.display = 'none';
             }
            break;
    }
}

function updateOrientationCalArea(message) {
    const orientationCalArea = document.getElementById("orientationCalAreaHelpText");
    orientationCalArea.innerHTML = message;
}

function stopCalibration(status) {
    calibrationStatus.innerHTML = status;
    resetCalibrationState();
}


